from car_framework.util import check_for_error, BATCH_SIZE
from car_framework.context import context


class BaseImport(object):

    def __init__(self):
        self.statuses = []

    def create_source_report_object(self):
        raise NotImplementedError()

    def wait_for_completion_of_import_jobs(self):
        context().car_service.check_import_status(self.statuses)
        for status in self.statuses:
            check_for_error(status)
        self.statuses = []

    def send_data(self, name, data):
        envelope = self.create_source_report_object()
        if name:
            envelope[name] = data
        status = context().car_service.import_data(envelope)
        check_for_error(status)
        self.statuses.append(status)
        if len(self.statuses) == BATCH_SIZE:
            self.wait_for_completion_of_import_jobs()

    def send_data_from_file(self, name, data_file):
        status = context().car_service.import_data_from_file(data_file)
        check_for_error(status)
        self.statuses.append(status)
        if len(self.statuses) == BATCH_SIZE:
            self.wait_for_completion_of_import_jobs()

    def get_last_model_state_id(self):
        return context().car_service.get_model_state_id()

    def save_new_model_state_id(self, new_model_state_id):
        return context().car_service.save_model_state_id(new_model_state_id)



import argparse, traceback, sys, os

from car_framework.context import Context, context
from car_framework.util import IncrementalImportNotPossible, RecoverableFailure, UnrecoverableFailure, DatasourceFailure


class BaseApp(object):
    def __init__(self, description):
        self.parser = argparse.ArgumentParser(description=description)
        self.parser.add_argument('-car-service-url', dest='car_service_apikey_url', default=os.getenv('CAR_SERVICE_URL',None), type=str, required=False, help='URL of the CAR ingestion service if API key is used for authorization')
        self.parser.add_argument('-car-service-key', dest='api_key', default=os.getenv('CAR_SERVICE_KEY',None), type=str, required=False, help='API key for CAR ingestion service')
        self.parser.add_argument('-car-service-password', dest='api_password', default=os.getenv('CAR_SERVICE_PASSWORD',None), type=str, required=False, help='Password for CAR ingestion service')

        self.parser.add_argument('-car-service-url-for-token', dest='car_service_token_url', default=os.getenv('CAR_SERVICE_URL_FOR_AUTHTOKEN',None), type=str, required=False, help='URL of the CAR ingestion service if Auth token is used for authorization')
        self.parser.add_argument('-car-service-token', dest='api_token', default=os.getenv('CAR_SERVICE_AUTHTOKEN',None), type=str, required=False, help='Auth token for CAR ingestion service')

        # source id to uniquely identify each data source
        self.parser.add_argument('-source', dest='source', default=os.getenv('CONNECTION_NAME',None), type=str, required=False, help='Unique source id for the data source')
        self.parser.add_argument('-name', dest='connector_name', default=os.getenv('CONNECTOR_NAME', None), type=str, required=False, help='Name of the connector')
        self.parser.add_argument('-version', dest='version', default=os.getenv('CONNECTOR_VERSION', None), type=str, required=False, help='Connector version number')

        self.parser.add_argument('-d', dest='debug', action='store_true', default=os.getenv('DEBUG', False), help='Enables DEBUG level logging')
        self.parser.add_argument('-connection-test', dest='connection_test', action='store_true', default=os.getenv('DATASOURCE_CONNECTION_TEST', False), help='Only perform datasource connection test and exit')
        self.parser.add_argument('-export-data-dir', dest='export_data_dir', default='/tmp/car_temp_export_data', help='Export data directory path, deafualt /tmp/car_temp_export_data')
        self.parser.add_argument('-keep-export-data-dir', dest='keep_export_data_dir', action='store_true', help='True for not removing export_data directory after complete, default false')
        self.parser.add_argument('-export-data-page-size', dest='export_data_page_size', type=int, default=2000, help='File export_data dump page size, default 2000')


    def setup(self):
        args = self.parser.parse_args()
        self.args = args

        if not args.api_token:
            if not args.api_key or not args.api_password:
                self.parser.print_usage(sys.stderr)
                sys.stderr.write('Either -car-service-token or -car-service-key and -car-service-password arguments are required.')
                sys.exit(2)

        if not args.car_service_apikey_url and not args.car_service_token_url:
            self.parser.print_usage(sys.stderr)
            sys.stderr.write('Either -car-service-url or -car-service-url-for-token is required.')
            sys.exit(2)

        if args.car_service_apikey_url:
            if not args.api_key or not args.api_password:
                self.parser.print_usage(sys.stderr)
                sys.stderr.write('If -car-service-url is provided then -car-service-key and -car-service-password arguments are required.')
                sys.exit(2)

        if args.car_service_token_url:
            if not args.api_token:
                self.parser.print_usage(sys.stderr)
                sys.stderr.write('If -car-service-url-for-token is provided then -car-service-token argument is required.')
                sys.exit(2)

        if not args.source:
            self.parser.print_usage(sys.stderr)
            sys.stderr.write('Missing required -source argument.')
            sys.exit(2)

        Context(args)


    def run(self):
        try:
            if self.args.connection_test:
                if hasattr(context(), 'asset_server') and hasattr(context().asset_server, 'test_connection') :
                    context().logger.info('Testing the datasource connection ... ')
                    code = context().asset_server.test_connection()
                    if code == 0:
                        context().logger.info('Testing the datasource connection was successful.')
                    else:
                        context().logger.error('Testing the datasource connection failed with code ' + str(code))
                    exit(code)
                else:
                    raise DatasourceFailure("The connector did not implement connection_test call.")
            else:
                try:
                    extension = self.get_schema_extension()
                    if extension: extension.setup()

                    context().logger.info('Attempting incremental import...')
                    context().inc_importer.run()
                except IncrementalImportNotPossible as e:
                    context().logger.info('Attempting full import...')
                    context().full_importer.run()

            context().logger.info('Done.')

        except RecoverableFailure as e:
            context().logger.info('Recoverable failure: ' + e.message)
            context().logger.info('Incremental import will be attempted again in the next run.')
            sys.exit(1)
        except UnrecoverableFailure as e:
            context().logger.info('Unrecoverable failure: ' + e.message)
            context().logger.info('Incremental import will not be possible in the next run.')
            context().car_service.reset_model_state_id()
            sys.exit(1)
        except DatasourceFailure as e:
            context().logger.info('Datasource failure: ' + e.message)
            sys.exit(1)
        except Exception as e:
            context().logger.exception(e)
            context().logger.error(traceback.format_exc())
            # traceback.print_exc()
            sys.exit(1)


    def get_schema_extension(self):
        return None



import logging
from pythonjsonlogger import jsonlogger
from datetime import datetime

class CustomJsonFormatter(jsonlogger.JsonFormatter):
    def add_fields(self, log_record, record, message_dict):
        # use inherited constructor
        super(CustomJsonFormatter, self).add_fields(log_record, record, message_dict)

        if not log_record.get('ibm_datetime'):
            now = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
            log_record['ibm_datetime'] = now
        
        if not log_record.get('connector') and context().args.connector_name:
            log_record['connector'] = context().args.connector_name
        
        if not log_record.get('source') and context().args.source:
            log_record['source'] = context().args.source
        
        if not log_record.get('version') and context().args.version:
            log_record['version'] = context().args.version

        # assign values to log_record
        log_record['level'] = log_record['level'].lower() if log_record.get('level') else record.levelname
        log_record['message'] = log_record['message'] if log_record.get('log') else record.message
        log_record['label'] = log_record['label'] if log_record.get('type') else record.name

def create_logger(debug = False, connector_name='Undefined'):
    logger = logging.getLogger()
    logger.setLevel(debug and logging.DEBUG or logging.INFO)

    handler = logging.StreamHandler()
    handler.setLevel(debug and logging.DEBUG or logging.INFO)
    format_string = '%(ibm_datetime)s %(level)s %(label)s %(message)s'
    formatter = CustomJsonFormatter(format_string)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger

def read_config(file_path, args):
    import json 

    try:
        with open(file_path) as f:
            config_file = json.load(f)
            args.connector_name = config_file.get('connection', {}).get('type', {}).get('displayName', {})
    except Exception:
        args.connector_name = ""


class Context(object):
    def __init__(self, args):
        global global_context
        global_context = self

        from car_framework.car_service import CarService
        from car_framework.communicator import Communicator
        self.args = args
        if not args.connector_name:
            read_config('configurations/config.json', self.args)
        self.logger = create_logger(args.debug, args.connector_name)
        self.car_service = CarService(Communicator())
        

global_context = None
def context():
    return global_context



from datetime import datetime
import json
import os
import shutil
import uuid

from car_framework.context import context
from car_framework.full_import import BaseFullImport


def get_report_time():
    delta = datetime.utcnow() - datetime(1970, 1, 1)
    milliseconds = delta.total_seconds() * 1000
    return milliseconds


class BaseDataHandler():

    source = None
    report = None
    source_report = None
    collections = {}
    collection_keys = {}
    edges = {}
    edge_keys = {}

    def __init__(self):
        self.timestamp = get_report_time()
        self.export_data_dir = os.path.join(context().args.export_data_dir, datetime.now().strftime('%Y-%m-%d_%H:%M:%S_r%f'))

    def create_source_report_object(self):
        raise NotImplementedError()

        # Adds the collection data
    def add_collection(self, name, object, key):
        objects = self.collections.get(name)
        if not objects:
            objects = []
            self.collections[name] = objects

        keys = self.collection_keys.get(name)
        if not keys:
            keys = []
            self.collection_keys[name] = keys

        if not object[key] in self.collection_keys[name]:
            objects.append(object)
            self.collection_keys[name].append(object[key])

        # dump collection to file to free memory
        if len(self.collections[name]) >= context().args.export_data_page_size:
            self._save_export_data_file(name, self.collections[name])
            self.collections[name] = []

    # Adds the edge between two vertices

    def add_edge(self, name, object):
        objects = self.edges.get(name)
        if not objects:
            objects = []
            self.edges[name] = objects

        keys = self.edge_keys.get(name)
        if not keys:
            keys = []
            self.edge_keys[name] = keys

        key = '#'.join(str(x) for x in object.values())
        if not key in self.edge_keys[name]:
            object['report'] = self.report['_key']
            object['source'] = context().args.source
            object['active'] = True
            object['timestamp'] = self.report['timestamp']
            objects.append(object)
            self.edge_keys[name].append(key)

        # dump edges to file to free memory
        if len(self.edges[name]) >= context().args.export_data_page_size:
            self._save_export_data_file(name, self.edges[name])
            self.edges[name] = []

    def send_collections(self, importer):
        context().logger.info('Creating vertices')
        if not self.collections and isinstance(importer, BaseFullImport):
            self._send_empty_report(importer)
        for name, data in self.collections.items():
            # save residual data
            if len(data) > 0:
                self._save_export_data_file(name, data)
            self._send(name, importer)
        context().logger.info('Creating vertices: done %s', {key: len(value) for key, value in self.collection_keys.items()})

    def send_edges(self, importer):
        context().logger.info('Creating edges')
        if not self.edges and isinstance(importer, BaseFullImport):
            self._send_empty_report(importer)
        for name, data in self.edges.items():
            # save residual data
            if len(data) > 0:
                self._save_export_data_file(name, data)
            self._send(name, importer)
        context().logger.info('Creating edges done: %s', {key: len(value) for key, value in self.edge_keys.items()})

    def _send_empty_report(self, iporter):
        iporter.send_data(None, None)

    def _create_export_data_dir(self, name):
        dir_path = os.path.join(self.export_data_dir, name)
        if not os.path.exists(dir_path):
            context().logger.debug('Creating export_data dir: %s', dir_path)
            os.makedirs(dir_path)

        return dir_path

    def _delete_export_data_dir(self, export_data_dir):
        if not context().args.keep_export_data_dir and os.path.exists(export_data_dir):
            context().logger.debug('Delete export_data dir: %s', export_data_dir)
            shutil.rmtree(export_data_dir)

    def _save_export_data_file(self, name, data):
        dir_path = self._create_export_data_dir(name)
        envelope = self.create_source_report_object()
        envelope[name] = data
        filename = os.path.join(dir_path, '%s.json' % str(uuid.uuid4())[0:8])
        with open(filename, 'w') as outfile:
            json.dump(envelope, outfile)
        return filename

    def _send(self, name, importer):
        dir_path = os.path.join(self.export_data_dir, name)
        for _, _, files in os.walk(dir_path):
            for data_file in files:
                with open(os.path.join(dir_path, data_file), 'r') as outfile:
                    importer.send_data_from_file(name, outfile)
        self._delete_export_data_dir(dir_path)

    def printData(self):
        context().logger.debug("Vertexes to be created:")
        context().logger.debug(self.collections)
        context().logger.debug("Edges to be created:")
        context().logger.debug(self.edges)



import setuptools
import os
import sys

version = os.environ.get('PYPI_PACKAGE_VERSION')
repository = os.environ.get('PYPI_PACKAGE_REPOSITORY')

with open("README.md", "r") as fh:
    long_description = fh.read()

#Prepare requires list
install_requires_list = set()
requirements_file = 'requirements-test.txt' if repository == 'test' else 'requirements.txt'
with open(requirements_file) as f:
    lines = f.readlines()
lines = [x.strip() for x in lines]
lines = list(filter(lambda s: (not s.startswith('#')) and len(s)>0, lines))
lines = list(filter(lambda s: (not s.startswith('git+')) and len(s)>0, lines))
install_requires_list.update(lines)
install_requires_list = list(install_requires_list)

setuptools.setup(
    name="car-connector-framework",
    version=version,
    author="IBM",
    author_email="",
    description="CAR service connector framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/IBM/cp4s-car-connector-framework",
    packages=setuptools.find_packages(exclude=['tests']),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=install_requires_list,
    license='Apache License 2.0',
    platforms=["Any"]
)